module.exports = class Bot {
    constructor() {
        var User = require('./user')
        var Command = require('./command');
        var Message = require('./message')
        var SocketIoEvent = require('./socketioevent')
        var getRoomData = require('../utils/roomlistparser')
        var he = require('he')
        var io = require('socket.io-client');
        var tbh = require('trollbox-headers').headers();
        
        this.socket = io("http://www.windows93.net:8081",tbh);
        this.name = "Trollbox Bot"
        this.color = "lime"
        this.users = {};
        this.rawusers = {};
        this._callbacks = {};
        this.__use_callbacks = [];

        this.commands = new Map();
        this.addons = new Map();

        this.cmdsfolder = __dirname+"/cmds";
        this.addonsfolder = __dirname+"/addons" /* both are default and common folders */
        this.prefix = "!" // default, common prefix
        this.room = "atrium"  // default room

        // Commands Variable Setup

        this.socket.on('message', (data) => {
            data.msg = he.decode(data.msg);

            this.commands.forEach((callback, cmd) => {
                if (typeof callback != "function") throw "Invalid Callback";
                var argsreq = callback.toString().includes('/* [COMMAND ARGUMENTS REQUIRED] */')
                if (data.msg.startsWith(argsreq ? `${this.prefix}${cmd} ` : `${this.prefix}${cmd}`)) {
                    callback(new Command(data,cmd,this.prefix,this.send),this);

                }
            })
            // Socket.io Event Setup
            this.__use_callbacks.forEach((f) => f(new SocketIoEvent("message",data), this))
            if (data.home == "trollbox" && data.msg.includes('entered')) {
                var d = getRoomData(data.msg)
                if (this._callbacks['join room'] != null) {
                    return this._callbacks['join room'].forEach((c) => c(d));

                }
            }
            if (this._callbacks.message != null) {
                this._callbacks.message.forEach((c) => c(new Message(data)))
            }
        })

        this.socket.on('user joined', (data) => {
            this.__use_callbacks.forEach((f) => f(new SocketIoEvent("user joined",data), this))
            
            if (this._callbacks.join != null) {
                this._callbacks.join.forEach((c) => c(new User(data)))
            }
        })
        this.socket.on('user left', (data) => {
            this.__use_callbacks.forEach((f) => f(new SocketIoEvent("user left",data), this))
            
            if (this._callbacks.left != null) {
                this._callbacks.left.forEach((c) => c(new User(data)))
            }
        });
        this.socket.on('update users', (data) => {
            this.rawusers = data;
                for (var key in data) {
                  if (!this.users[data[key].home]) {
                    this.users[data[key].home]=[he.decode(data[key].nick)]
                    
                  }else{
                    this.users[data[key].home].push(he.decode(data[key].nick));
                  }
                }
                
              
        });
        this.socket.on('disconnect',(reason) => {
            this.triggerEvent('disconnect',reason)
        })
        this.socket.on('_connected',() => {
            this.triggerEvent('connect',null)
        })
        

    }
    triggerEvent(event,data) {
        this.__use_callbacks.forEach((f) => f(new SocketIoEvent("user joined",data), this))
        
        if (this._callbacks[event] != null) {
            this._callbacks[event].forEach((c) => c(new User(data)))
        }
    }
    setName(name) {
        this.name = name;
    }
    setColor(color) {
        this.color = color;

    }
    changeColor(color) {
        this.setColor(color);
        this.socket.emit('user joined',this.name,this.color,"","");
    }
    setPrefix(prefix) {
        this.prefix = prefix;
    }
    connect() {
        if (!this.socket.connected) this.socket.connect();
        this.socket.emit('user joined',this.name,this.color,"","")
        
    }
    disconnect() {
        if (!this.socket.connected) throw "You need to be connected to disconnect."
        this.socket.disconnect()

    }
    on(event,callback) {
        if (this._callbacks[event] == null) this._callbacks[event] = [];
        this._callbacks[event].push(callback);
    }
    sio_once(event,callback) {
        this.socket.once(event,(...data) => {
            this.callback(this,...data)
        })
    }
    use(func) {
        if (typeof func != "function") throw "This is not a function!"
        this.__use_callbacks.push(func);
    }
    send(msg,disableAntiSpamBypass = false) {
        this.socket.send(msg+(disableAntiSpamBypass ? "" : " ".repeat(Math.floor(Math.random() * 15))))
    }
    waitForMessage(msg,clk) {
        var func = (d) => {
            if (d.msg.includes(msg)) {
                clk(d)
            } else {
                this.socket.once('message',func)
            }
        }
        this.socket.once('message',func)
    }
    waitForHome(home,clk) {
        var func = (d) => {
            if (d.home.includes(home)) {
                clk(d)
            } else {
                this.socket.once('message',func)
            }
        }
        this.socket.once('message',func)
    }
    setRoom(room) {
        this.send(`/r ${room}`,1);
        this.room = room;

    }
    defaultRoom() {
        this.send(`/a`,1);
        this.room = "atrium";
    }
    getRoom() {
        return this.room;
    }
    getRooms() {
        var roomListParser = require('../utils/roomlistparser');
        var res = roomListParser(this.rawusers)
        return res;
        
        
    }
    setCommandsFolder(path) {
        this.cmdsfolder = path;
        this.reloadCmdsAndAddons();
    }
    setAddonsFolder(path) {
        this.addonsfolder = path;
        this.reloadCmdsAndAddons();
    }
    reloadCmdsAndAddons() {
        // deletes all CommonJS Cache Data
        for (var item in require.cache) {
            delete require.cache[item];
        }
        // now this is the fun part
        var fs = require('fs');
        if (fs.existsSync(this.cmdsfolder) && fs.existsSync(this.addonsfolder)) {
            fs.readdir(this.cmdsfolder,(err,files) => {
                for (i = 0; i < files.length; i++) {
                    loadCommandFromFile(require(files[i]))
                }
            })
            fs.readdir(this.addonsfolder,(err,files) => {
                for (i = 0; i < files.length; i++) {
                    loadAddonFromFile(require(files[i]))
                }
            })
            
        }
    }
    loadCommandFromFile(file) {
        var info = file.main({
            rawjson: {
                metadata: {}
            },
            setCommandName(name) {
                this.rawjson.name = name;
            },
            // setCommandDescription(description) {
            //     this.rawjson.description = description;
            // },
            setCommandFunction(func) {
                this.rawjson.func = func;
            },
            setMetadata(meta) {
                this.rawjson.metadata = meta;
            },
            done() {
                return this.rawjson;
            }
        });
        if (!info.name && !info.func) throw "Invalid Command Missing Function or Name"
        this.commands.set(info.name,info.func)
    }
    loadAddonFromFile(file) {
        var botc = this;
        var info = file.premain({
            rawjson: {
                metadata: {}
            },
            setAddonName(name) {
                this.rawjson.name = name;
            },
            setAddonID(id) {
                this.rawjson.id = id;
            },
            setAddonMainFunction(func) {
                this.rawjson.func = func;
            },
            setMetadata(meta) {
                this.rawjson.metadata = meta;
            }
        },function() {
            if (!info.name && !info.func && !info.id) throw "Invalid Addon Missing Function, ID or Name"
            this.addons.set(info.id,info.rawjson);
            var addon = this.addons.get(info.id);
            
            addon.func({
                listen(event,callback) {
                    this.on(event,callback)
                },
                getEvent(event) {
                    return new Promise((resolve) => {
                        this.sio_once(event,(...data) => {
                            resolve(...data)
                        })
                    })
                },
                getBot() {
                    return botc;
                },
                invokeBotFunction(name,...params) {
                    return botc[name](...params);
                }
            })
        });
        
    }
    
    
}